# implementation of <<An integrated allocation method for the PCB assembly line balancing problem with nozzle changes>>
from base_optimizer.optimizer_common import *
from optimizer_hyperheuristic import *


def selective_initialization(component_points, component_feeders, population_size, machine_number):
    population = []    # population initialization
    for _ in range(population_size):
        individual = []
        for part_index, points in component_points:
            if points == 0:
                continue
            # 可用机器数
            avl_machine_num = random.randint(1, min(machine_number, component_feeders[part_index], points))

            selective_possibility = []
            for p in range(1, avl_machine_num + 1):
                selective_possibility.append(pow(2, avl_machine_num - p + 1))

            sel_machine_num = random_selective([p + 1 for p in range(avl_machine_num)], selective_possibility)  # 选择的机器数
            sel_machine_set = random.sample([p for p in range(machine_number)], sel_machine_num)

            sel_machine_points = [1 for _ in range(sel_machine_num)]
            for p in range(sel_machine_num - 1):
                if points == sum(sel_machine_points):
                    break
                assign_points = random.randint(1, points - sum(sel_machine_points))
                sel_machine_points[p] += assign_points

            if sum(sel_machine_points) < points:
                sel_machine_points[-1] += (points - sum(sel_machine_points))

            # code component allocation into chromosome
            for p in range(machine_number):
                if p in sel_machine_set:
                    individual += [0 for _ in range(sel_machine_points[0])]
                    sel_machine_points.pop(0)
                individual.append(1)
            individual.pop(-1)

        population.append(individual)
    return population


def selective_crossover(component_points, component_feeders, mother, father, machine_number, non_decelerating=True):
    assert len(mother) == len(father)

    offspring1, offspring2 = mother.copy(), father.copy()
    one_counter, feasible_cut_line = 0, []

    idx = 0
    for part_index, points in component_points:
        one_counter = 0

        idx_, mother_cut_line, father_cut_line = 0, [-1], [-1]
        for idx_, gene in enumerate(mother[idx: idx + points + machine_number - 1]):
            if gene:
                mother_cut_line.append(idx_)
        mother_cut_line.append(idx_ + 1)

        for idx_, gene in enumerate(father[idx: idx + points + machine_number - 1]):
            if gene:
                father_cut_line.append(idx_)
        father_cut_line.append(idx_ + 1)

        for offset in range(points + machine_number - 1):
            if mother[idx + offset] == 1:
                one_counter += 1
            if father[idx + offset] == 1:
                one_counter -= 1

            # first constraint: the total number of '1's (the number of partitions) in the chromosome is unchanged
            if one_counter != 0 or offset == 0 or offset == points + machine_number - 2:
                continue

            # the selected cut-line should guarantee there are the same or a larger number unassigned machine
            # for each component type
            n_bro, n_new = 0, 0
            if mother[idx + offset] and mother[idx + offset + 1]:
                n_bro += 1
            if father[idx + offset] and father[idx + offset + 1]:
                n_bro += 1
            if mother[idx + offset] and father[idx + offset + 1]:
                n_new += 1
            if father[idx + offset] and mother[idx + offset + 1]:
                n_new += 1

            # second constraint: non_decelerating or accelerating crossover
            # non_decelerating or accelerating means that the number of machine without workload is increased
            if n_new < n_bro or (n_new == n_bro and not non_decelerating):
                continue

            # third constraint (customized constraint):
            # no more than the maximum number of available machine for each component type
            new_mother_cut_line, new_father_cut_line = [], []
            for idx_ in range(machine_number + 1):
                if mother_cut_line[idx_] <= offset:
                    new_mother_cut_line.append(mother_cut_line[idx_])
                else:
                    new_father_cut_line.append(mother_cut_line[idx_])

                if father_cut_line[idx_] <= offset:
                    new_father_cut_line.append(father_cut_line[idx_])
                else:
                    new_mother_cut_line.append(father_cut_line[idx_])

            sorted(new_mother_cut_line, reverse=False)
            sorted(new_father_cut_line, reverse=False)
            n_mother_machine, n_father_machine = 0, 0

            for idx_ in range(machine_number):
                if new_mother_cut_line[idx_ + 1] - new_mother_cut_line[idx_] > 1:
                    n_mother_machine += 1

                if new_father_cut_line[idx_ + 1] - new_father_cut_line[idx_] > 1:
                    n_father_machine += 1

            if n_mother_machine > component_feeders[part_index] or n_father_machine > component_feeders[part_index]:
                continue

            feasible_cut_line.append(idx + offset)

        idx += (points + machine_number - 1)

    if len(feasible_cut_line) == 0:
        return offspring1, offspring2

    cut_line_idx = feasible_cut_line[random.randint(0, len(feasible_cut_line) - 1)]
    offspring1, offspring2 = mother[:cut_line_idx + 1] + father[cut_line_idx + 1:], father[:cut_line_idx + 1] + mother[
                                                                                                     cut_line_idx + 1:]
    return offspring1, offspring2


def cal_individual_val(component_points, component_feeders, component_nozzle, machine_number, individual, data_mgr, net):
    idx, objective_val = 0, []
    machine_component_points = [[] for _ in range(machine_number)]
    nozzle_component_points = defaultdict(list)

    # decode the component allocation
    for comp_idx, points in component_points:
        component_gene = individual[idx: idx + points + machine_number - 1]
        machine_idx, component_counter = 0, 0
        for gene in component_gene:
            if gene:
                machine_component_points[machine_idx].append(component_counter)
                machine_idx += 1
                component_counter = 0
            else:
                component_counter += 1
        machine_component_points[-1].append(component_counter)
        idx += (points + machine_number - 1)

        nozzle_component_points[component_nozzle[comp_idx]] = [0] * len(component_points)   # 初始化元件-吸嘴点数列表

    # ======== 新加的开始 ========
    for machine_idx in range(machine_number):
        cp_points, cp_nozzle = defaultdict(int), defaultdict(str)
        for comp_idx, _ in component_points:
            if machine_component_points[machine_idx][comp_idx] == 0:
                continue
            cp_points['C' + str(comp_idx)] = machine_component_points[machine_idx][comp_idx]
            cp_nozzle['C' + str(comp_idx)] = component_nozzle[comp_idx]

        encoding = np.array(data_mgr.encode(cp_points, cp_nozzle, 45, 150))
        encoding = torch.from_numpy(encoding.reshape((-1, np.shape(encoding)[0]))).float().to("cuda")
        # pred_time = net(encoding)[0, 0].item()
        # objective_val.append(pred_time * sum(points for points in cp_points.values()))
        objective_val.append(net(encoding)[0, 0].item())

    return objective_val, machine_component_points
    # ======== 新加的结束（以下内容弃用） =====
    for comp_idx, points in component_points:
        nozzle_component_points[component_nozzle[comp_idx]][comp_idx] = points

    for machine_idx in range(machine_number):
        nozzle_points = defaultdict(int)
        for idx, nozzle in component_nozzle.items():
            if component_points[idx] == 0:
                continue
            nozzle_points[nozzle] += machine_component_points[machine_idx][idx]

        machine_points = sum(machine_component_points[machine_idx])         # num of placement points
        if machine_points == 0:
            continue
        ul = math.ceil(len(nozzle_points) * 1.0 / max_head_index) - 1       # num of nozzle set

        # assignments of nozzles to heads
        wl = 0                                                          # num of workload
        total_heads = (1 + ul) * max_head_index - len(nozzle_points)
        nozzle_heads = defaultdict(int)
        for nozzle in nozzle_points.keys():
            if nozzle_points[nozzle] == 0:
                continue
            nozzle_heads[nozzle] = math.floor(nozzle_points[nozzle] * 1.0 / machine_points * total_heads)
            nozzle_heads[nozzle] += 1

        total_heads = (1 + ul) * max_head_index
        for heads in nozzle_heads.values():
            total_heads -= heads

        while True:
            nozzle = max(nozzle_heads, key=lambda x: nozzle_points[x] / nozzle_heads[x])
            if total_heads == 0:
                break
            nozzle_heads[nozzle] += 1
            total_heads -= 1

        # averagely assign placements to heads
        heads_placement = []
        for nozzle in nozzle_heads.keys():
            points = math.floor(nozzle_points[nozzle] / nozzle_heads[nozzle])

            heads_placement += [[nozzle, points] for _ in range(nozzle_heads[nozzle])]
            nozzle_points[nozzle] -= (nozzle_heads[nozzle] * points)
            for idx in range(len(heads_placement) - 1, -1, -1):
                if nozzle_points[nozzle] <= 0:
                    break
                nozzle_points[nozzle] -= 1
                heads_placement[idx][1] += 1
        heads_placement = sorted(heads_placement, key=lambda x: x[1], reverse=True)

        # the number of pick-up operations
        # (under the assumption of the number of feeder available for each comp. type is equal 1)
        pl = 0
        heads_placement_points = [0 for _ in range(max_head_index)]
        while True:
            head_assign_point = []
            for head in range(max_head_index):
                if heads_placement_points[head] != 0 or heads_placement[head] == 0:
                    continue

                nozzle, points = heads_placement[head]
                max_comp_index = np.argmax(nozzle_component_points[nozzle])

                heads_placement_points[head] = min(points, nozzle_component_points[nozzle][max_comp_index])
                nozzle_component_points[nozzle][max_comp_index] -= heads_placement_points[head]

                head_assign_point.append(heads_placement_points[head])

            min_points_list = list(filter(lambda x: x > 0, heads_placement_points))
            if len(min_points_list) == 0 or len(head_assign_point) == 0:
                break

            pl += max(head_assign_point)

            for head in range(max_head_index):
                heads_placement[head][1] -= min(min_points_list)
                heads_placement_points[head] -= min(min_points_list)

        # every max_head_index heads in the non-decreasing order are grouped together as nozzle set
        for idx in range(len(heads_placement) // max_head_index):
            wl += heads_placement[idx][1]
        objective_val.append(T_pp * machine_points + T_tr * wl + T_nc * ul + T_pl * pl)

    return objective_val, machine_component_points


def individual_convert(component_points, individual):
    machine_number = len(individual)
    machine_component_points = [[] for _ in range(machine_number)]
    idx = 0
    # decode the component allocation
    for comp_idx, points in component_points:
        component_gene = individual[idx: idx + points + machine_number - 1]
        machine_idx, component_counter = 0, 0
        for gene in component_gene:
            if gene:
                machine_component_points[machine_idx].append(component_counter)
                machine_idx += 1
                component_counter = 0
            else:
                component_counter += 1
        machine_component_points[-1].append(component_counter)
        idx += (points + machine_number - 1)

    return machine_component_points


def assemblyline_optimizer_genetic(pcb_data, component_data, machine_number):
    # basic parameter
    # crossover rate & mutation rate: 80% & 10%
    # population size: 200
    # the number of generation: 500
    crossover_rate, mutation_rate = 0.8, 0.1
    population_size, n_generations = 200, 500
    # population_size, n_generations = 30, 50

    # the number of placement points, the number of available feeders, and nozzle type of component respectively
    component_points, component_feeders, component_nozzle = defaultdict(int), defaultdict(int), defaultdict(str)
    for data in pcb_data.iterrows():
        part_index = component_data[component_data['part'] == data[1]['part']].index.tolist()[0]
        nozzle = component_data.loc[part_index]['nz']

        component_points[part_index] += 1
        component_feeders[part_index] = component_data.loc[part_index]['feeder-limit']
        component_nozzle[part_index] = nozzle

    component_points = sorted(component_points.items(), key=lambda x: x[0])     # 决定染色体排列顺序
    data_mgr = DataMgr()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    net = Net(input_size=data_mgr.get_feature(), output_size=1).to(device)

    net.load_state_dict(torch.load('model_state.pth'))
    # optimizer = torch.optim.Adam(net.parameters(), lr=0.1)
    # optimizer.load_state_dict(torch.load('optimizer_state.pth'))

    # population initialization
    population = selective_initialization(component_points, component_feeders, population_size, machine_number)
    with tqdm(total=n_generations) as pbar:
        pbar.set_description('genetic algorithm process for PCB assembly line balance')

        new_population = []
        for _ in range(n_generations):
            # calculate fitness value
            pop_val = []
            for individual in population:
                val, assigned_points = cal_individual_val(component_points, component_feeders, component_nozzle,
                                                          machine_number, individual, data_mgr, net)
                pop_val.append(max(val))

            select_index = get_top_k_value(pop_val, population_size - len(new_population), reverse=False)
            population = [population[idx] for idx in select_index]
            pop_val = [pop_val[idx] for idx in select_index]

            population += new_population
            for individual in new_population:
                val, _ = cal_individual_val(component_points, component_feeders, component_nozzle, machine_number,
                                            individual, data_mgr, net)
                pop_val.append(max(val))

            # min-max convert
            max_val = max(pop_val)
            pop_val = list(map(lambda v: max_val - v, pop_val))
            sum_pop_val = sum(pop_val) + 1e-10
            pop_val = [v / sum_pop_val + 1e-3 for v in pop_val]

            # crossover and mutation
            new_population = []
            for pop in range(population_size):
                if pop % 2 == 0 and np.random.random() < crossover_rate:
                    index1 = roulette_wheel_selection(pop_val)
                    while True:
                        index2 = roulette_wheel_selection(pop_val)
                        if index1 != index2:
                            break

                    offspring1, offspring2 = selective_crossover(component_points, component_feeders,
                                                                 population[index1], population[index2], machine_number)

                    if np.random.random() < mutation_rate:
                        offspring1 = constraint_swap_mutation(component_points, offspring1, machine_number)

                    if np.random.random() < mutation_rate:
                        offspring2 = constraint_swap_mutation(component_points, offspring2, machine_number)

                    new_population.append(offspring1)
                    new_population.append(offspring2)

            pbar.update(1)

    best_individual = population[np.argmax(pop_val)]
    val, assignment_result = cal_individual_val(component_points, component_feeders, component_nozzle, machine_number,
                                                best_individual, data_mgr, net)
    print('final value: ', val)
    # available feeder check
    for part_index, data in component_data.iterrows():
        feeder_limit = data['feeder-limit']
        for machine_index in range(machine_number):
            if assignment_result[machine_index][part_index]:
                feeder_limit -= 1
        assert feeder_limit >= 0

    return assignment_result
